// EEPROM access function for H8/3664N
//               by I.N.
// 2007/08/27 ver.1.0

// Memo: Need 10ms before read function after write function

struct st_iic {				// struct IIC 
	union {					// ICCR 
		unsigned char BYTE; // Byte Access 
		struct {			// Bit Access 
			unsigned char ICE :1; // ICE 
			unsigned char IEIC:1; // IEIC 
			unsigned char MST :1; // MST 
			unsigned char TRS :1; // TRS 
			unsigned char ACKE:1; // ACKE 
			unsigned char BBSY:1; // BBSY 
			unsigned char IRIC:1; // IRIC 
			unsigned char SCP :1; // SCP 
		} BIT;
	} _ICCR;

	union {					// ICSR 
		unsigned char BYTE; // Byte Access 
		struct {			// Bit Access 
			unsigned char ESTP:1; // ESTP 
			unsigned char STOP:1; // STOP 
			unsigned char IRTR:1; // IRTR 
			unsigned char AASX:1; // AASX 
			unsigned char AL  :1; // AL 
			unsigned char AAS :1; // AAS 
			unsigned char ADZ :1; // ADZ 
			unsigned char ACKB:1; // ACKB 
		} BIT;
	} _ICSR;

	union {
		struct {
			union {			// SARX 
				unsigned char BYTE; // Byte Access 
				struct {			// Bit Access 
					unsigned char SVAX:7; // SVAX 
					unsigned char FSX :1; // FSX 
				} BIT;
			} UN_SARX;
			union {			// SAR 
				unsigned char BYTE; // Byte Access 
				struct {			// Bit Access 
					unsigned char SVA:7; // SVA 
					unsigned char FS :1; // FS 
				} BIT;
			} UN_SAR;
		} ICE0;
		struct {
			unsigned char UN_ICDR;	// ICDR 
			union {					// ICMR 
				unsigned char BYTE; // Byte Access 
				struct {			// Bit Access 
					unsigned char MLS :1; // MLS 
					unsigned char WAIT:1; // WAIT 
					unsigned char CKS :3; // CKS 
					unsigned char BC  :3; // BC 
				} BIT;
			} UN_ICMR;
		} ICE1;
	} EQU;
};

union un_tscr {			// union TSCR 
	unsigned char BYTE; // Byte Access 
	struct {			// Bit Access 
		unsigned char wk    :6;
		unsigned char IICRST:1; // IICRST 
		unsigned char IICX  :1; // IICX 
	} BIT;
};

#define _IIC	(*(volatile struct st_iic *)0xFFC4)		// IIC Address 
#define _ICDR	EQU.ICE1.UN_ICDR						// ICDR Change 
#define _ICMR	EQU.ICE1.UN_ICMR						// ICDR Change 
#define _TSCR	(*(volatile union un_tscr *)0xFFFC)		// TSCR Address 

#define DEVICE_CODE		0xa0	// EEPROM DEVICE CODE:101
#define SLAVE_ADRS		0x00	// SLAVE ADRS:0
#define IIC_DATA_W		0x00	// WRITE_DATA
#define IIC_DATA_R		0x01	// READ_DAT

#define WR_RETRY_CNT	10		// tWC(1ms)*10=10ms
#define WR_OK			11		// loop break

// 100=100us
void wait_eeprom(unsigned int loop)
{
	unsigned int i;
	unsigned char dummy;
	for(i=0; i<loop; i++){
		dummy++;
	}
}

//----------------------------------------------------------------------
//	Set_adrs_EEPROM
//	Argument1: Write / read address (unsigned short)
//	Return value: 1: OK / 0: NG EEPROM NOACK(unsigned char)
//----------------------------------------------------------------------
//----------------------------------------------------------------------
//  ADDRESS SET ACTION / DUMMY WRITE ACTION)
//  <1>			<2> 			<3>
//  123456789 	123456789 		123456789
//  101000000 	000000000 		000000000
//  Slave WA 	Address HI A 	Address LO A
//----------------------------------------------------------------------
unsigned char Set_adrs_EEPROM(unsigned short adrs)
{
	unsigned char ret;

	_IIC._ICCR.BYTE |= 0x30;			// Master transmission setting (MST=1,TRS=1) 
	_IIC._ICCR.BYTE = ((_IIC._ICCR.BYTE & 0xfe) | 0x04);
	// Generating of start condition 
	while (_IIC._ICCR.BIT.IRIC == 0);			// Transmission OK? 
	// <1> 
	_IIC._ICDR = (unsigned char)(DEVICE_CODE | SLAVE_ADRS | IIC_DATA_W);
	// <1>Slave address set 	
	_IIC._ICCR.BIT.IRIC = 0;
	while (_IIC._ICCR.BIT.IRIC == 0);			// <1> Transmission complete? 
	if (_IIC._ICSR.BIT.ACKB != 0) return 0;		// ACK? 
	// <2> 
	_IIC._ICDR = (unsigned char)(adrs >> 8);	// <2> Upper address set 
	_IIC._ICCR.BIT.IRIC = 0;
	while (_IIC._ICCR.BIT.IRIC == 0);			// <2> Transmission complete? 
	if (_IIC._ICSR.BIT.ACKB != 0) return 0;		// ACK? 
	// <3> 
	_IIC._ICDR = (unsigned char)(adrs & 0x00ff);	// <3> Lower address set 
	_IIC._ICCR.BIT.IRIC = 0;
	while (_IIC._ICCR.BIT.IRIC == 0);			// <3> Transmission complete? 
	if (_IIC._ICSR.BIT.ACKB != 0) return 0;		// ACK? 
	return 1;
}

//----------------------------------------------------------------------
//	Recv_datan_EEPROM
//	Argument1: Storing read data address (unsigned char *)
//	Argument2: Read byte number1-512 (unsigned char)
//	Return value: 1: OK / 0: NG EEPROM NOACK (unsigned char)
//----------------------------------------------------------------------
//----------------------------------------------------------------------
//  (CURRENT ADDRESS READ ACTION)
//  <1>			<2>				<3>				<n>
//  123456789 	123456789 		123456789   	123456789
//  101000010 	000000000 		000000000   	000000000
//  Slave RA 	Read data A 	Read data A		Read data A
//----------------------------------------------------------------------
unsigned char Recv_datan_EEPROM(unsigned char *rd_ptr, unsigned short no)
{
	unsigned char recv;

	_IIC._ICCR.BYTE |= 0x30;		// Master transmission setting (MST=1,TRS=1) 
	_IIC._ICCR.BYTE = ((_IIC._ICCR.BYTE & 0xfe) | 0x04);
	// Generating of start condition 
	while (_IIC._ICCR.BIT.IRIC == 0);			// Transmission OK? 
	_IIC._ICDR = (unsigned char)(DEVICE_CODE | SLAVE_ADRS | IIC_DATA_R);
	// <1> Slave address set 
	_IIC._ICCR.BIT.IRIC = 0;
	while (_IIC._ICCR.BIT.IRIC == 0);			// <1> Transmission complete? 
	if ( _IIC._ICSR.BIT.ACKB != 0 ) return 0;	// ACK? 
	wait_eeprom(30);		// dummy wait 
	// Master reception setting 
	_IIC._ICCR.BIT.TRS = 0;
	_IIC._ICMR.BIT.WAIT = 1;
	_IIC._ICSR.BIT.ACKB = 0;
	recv = _IIC._ICDR;							// dummy read 
	_IIC._ICCR.BIT.IRIC = 0;
	while (_IIC._ICCR.BIT.IRIC == 0);			// Reception complete? 
	_IIC._ICCR.BIT.IRIC = 0;
	while(1<no){
		while (_IIC._ICCR.BIT.IRIC == 0);		// Reception complete? 
		*rd_ptr = _IIC._ICDR;					// Received data read 
		_IIC._ICCR.BIT.IRIC = 0;
		while (_IIC._ICCR.BIT.IRIC == 0);		// Reception complete?	
		rd_ptr++;								// Storage address increment 
		no--;
		if(1<no) _IIC._ICCR.BIT.IRIC = 0;
	}
	// < The last reception > 
	_IIC._ICSR.BIT.ACKB = 1;
	_IIC._ICCR.BIT.TRS = 1;
	_IIC._ICCR.BIT.IRIC = 0;
	while (_IIC._ICCR.BIT.IRIC == 0);			// Reception complete? 
	_IIC._ICMR.BIT.WAIT = 0;
	*rd_ptr = _IIC._ICDR;						// Received data read 
	_IIC._ICCR.BIT.IRIC = 0;
	_IIC._ICCR.BYTE = _IIC._ICCR.BYTE & 0xfa;	// Generation of stop condition 
	return 1;
}

//----------------------------------------------------------------------
//	Read_page_EEPROM	(8byte)
//	Argument1: Read address(unsigned short)
//	Argument2: Storing read data address (unsigned char *)
//	Return value: 1: OK / 0: NG EEPROM NOACK(unsigned char)
//----------------------------------------------------------------------
unsigned char Read_page_EEPROM(unsigned short adrs, unsigned char *rd_ptr)
{
	unsigned char i,ack;

	_IIC._ICCR.BYTE = 0x89;				// ICE=1 P57,P56->SCL,SDA 
	_IIC._ICMR.BYTE = 0x08;
	_TSCR.BYTE = 0xfc;
	ack = 0;
	while (_IIC._ICCR.BIT.BBSY != 0);	// Bus busy? 
	ack = Set_adrs_EEPROM(adrs);		// Addressing(dummy write) 
	if (ack == 0)	{
		_IIC._ICCR.BIT.IRIC = 0;
		_IIC._ICCR.BYTE = _IIC._ICCR.BYTE & 0xfa;	// Generate stop condition 
		return 0;
	}
	ack = Recv_datan_EEPROM(rd_ptr,8);	// Data reading of 8 bytes 
	return ack;
}

//----------------------------------------------------------------------
//	Read_page_EEPROM	(8byte)
//	Argument1: Read address (unsigned short)
//	Argument2: Storing read data address (unsigned char *)
//	Argument3: Number of read data (unsigned short)
//	Return value: 1: OK / 0: NG EEPROM NOACK(unsigned char)
//----------------------------------------------------------------------
unsigned char Read_n_EEPROM(unsigned short adrs, unsigned char *rd_ptr, unsigned short no)
{
	unsigned char	i,ack;

	_IIC._ICCR.BYTE = 0x89;				// ICE=1 P57,P56->SCL,SDA 
	_IIC._ICMR.BYTE = 0x08;
	_TSCR.BYTE = 0xfc;
	ack = 0;
	while (_IIC._ICCR.BIT.BBSY != 0);	// Bus busy? 
	ack = Set_adrs_EEPROM(adrs);		// Addressing (dummy write) 
	if (ack==0){
		_IIC._ICCR.BIT.IRIC = 0;
		_IIC._ICCR.BYTE = _IIC._ICCR.BYTE & 0xfa;	// Generate stop condition 
		return 0;
	}
	ack = Recv_datan_EEPROM(rd_ptr,no);	// Data reading of n bytes 
	return ack;
}

//----------------------------------------------------------------------
//	Recv_data1_EEPROM
//	Return value: Read data (unsigned char)
//----------------------------------------------------------------------
//----------------------------------------------------------------------
//  (CURRENT ADDRESS READ ACTION)
//  <1>			<2>
//  123456789 	123456789
//  101000010 	000000000
//  Slave RA 	Read data A
//----------------------------------------------------------------------
unsigned char Recv_data1_EEPROM(void)
{
	unsigned char recv;

	_IIC._ICCR.BYTE |= 0x30;	// Master transmission setting (MST=1,TRS=1) 
	_IIC._ICCR.BYTE = ((_IIC._ICCR.BYTE & 0xfe) | 0x04);
	// Generating of start condition 	
	while (_IIC._ICCR.BIT.IRIC == 0);	// Transmission OK? 
	_IIC._ICDR = (unsigned char)(DEVICE_CODE | SLAVE_ADRS | IIC_DATA_R);
	// <1> Slave address set 
	_IIC._ICCR.BIT.IRIC = 0;
	while (_IIC._ICCR.BIT.IRIC == 0);	// <1> Transmission complete? 
	if ( _IIC._ICSR.BIT.ACKB != 0 ) return 0;	// ACK? 
	wait_eeprom(30);					// dummy wait 
	// Master reception setting 
	_IIC._ICCR.BIT.TRS = 0;
	_IIC._ICMR.BIT.WAIT = 1;
	_IIC._ICSR.BIT.ACKB = 0;
	recv = _IIC._ICDR;					// dummy read 
	_IIC._ICCR.BIT.IRIC = 0;
	while (_IIC._ICCR.BIT.IRIC == 0);	// Reception complete? 
	// < The last reception > 
	_IIC._ICSR.BIT.ACKB = 1;
	_IIC._ICCR.BIT.TRS = 1;
	_IIC._ICCR.BIT.IRIC = 0;
	while (_IIC._ICCR.BIT.IRIC == 0);	// Reception complete? 
	_IIC._ICMR.BIT.WAIT = 0;
	recv = _IIC._ICDR;					// Received data read 
	return recv;
}

//----------------------------------------------------------------------
//	Read_byte_EEPROM
//	Argument1: Read address (unsigned short)
//	Return value: Read data (unsigned char)
//----------------------------------------------------------------------
unsigned char Read_byte_EEPROM(unsigned short adrs)
{
	unsigned char	data,ack;

	_IIC._ICCR.BYTE = 0x89;					// ICE=1 P57,P56->SCL,SDA 
	_IIC._ICMR.BYTE = 0x08;
	_TSCR.BYTE = 0xfc;
	while (_IIC._ICCR.BIT.BBSY != 0);		// Bus busy? 
	ack = Set_adrs_EEPROM(adrs);			// Addressing (dummy write)
	if (ack == 1)	{
		data = Recv_data1_EEPROM();
	}
	_IIC._ICCR.BIT.IRIC = 0;
	_IIC._ICCR.BYTE = _IIC._ICCR.BYTE & 0xfa;	// Generate stop condition
	return data;
}

//----------------------------------------------------------------------
//	Write_data_EEPROM
//	Argument1: Write address (unsigned short)
//	Return value: 1: OK / 0: NG EEPROM NOACK(unsigned char)
//----------------------------------------------------------------------
//----------------------------------------------------------------------
//  (BYTE WRITE ACTION)
//	  <4>      		
//	  123456789		
//	  000000000		
//	  Write data A	
//----------------------------------------------------------------------
unsigned char Write_data_EEPROM(unsigned char wr_data)
{
	// <4> 
	_IIC._ICDR = wr_data;					// <4> Set writing data 
	_IIC._ICCR.BIT.IRIC = 0;
	while (_IIC._ICCR.BIT.IRIC == 0);		// <4> Transmission complete? 
	if ( _IIC._ICSR.BIT.ACKB != 0 ) return 0;	// ACK? 	
	return 1;
}

//----------------------------------------------------------------------
//	Write_page_EEPROM	(8byte)
//	Argument1: Write address (unsigned short)
//	Argument2: Storing write data address (unsigned char *)
//	Return value: 1: OK / 0: NG EEPROM NOACK(unsigned char)
//----------------------------------------------------------------------
unsigned char Write_page_EEPROM(unsigned short adrs ,unsigned char *wr_ptr)
{
	unsigned char	loop,i,ack;

	ack = 0;
	loop = 0;
	while (loop <= WR_RETRY_CNT){
		_IIC._ICCR.BYTE = 0x89;				// ICE=1(P57,P56->SCL,SDA) 
		_IIC._ICMR.BYTE = 0x08;
		_TSCR.BYTE = 0xfc;
		while (_IIC._ICCR.BIT.BBSY != 0);	// Bus busy? 
		ack = Set_adrs_EEPROM(adrs);		// Addressing 
		if (ack==1){
			loop = WR_OK;
		}else{
			_IIC._ICCR.BIT.IRIC = 0;
			_IIC._ICCR.BYTE = _IIC._ICCR.BYTE & 0xfa;	// Generation of stop condition 
			wait_eeprom(1000);				// tWC=1ms 
			loop++;
		}
	}
	if (ack==1){
		_IIC._ICCR.BIT.IRIC = 0;
		i = 0;
		while ((i<8) && (ack==1)){
			ack = Write_data_EEPROM(*wr_ptr);		// Writing of data of 8 bytes 
			wr_ptr++;
			i++;
		}
		_IIC._ICCR.BIT.IRIC = 0;
		_IIC._ICCR.BYTE = _IIC._ICCR.BYTE & 0xfa;	// Generation of stop condition 
	}
	return ack;
}

//----------------------------------------------------------------------
//	Write_byte_EEPROM
//	Argument1: Write address (unsigned short)
//	Argument2: Write data (unsigned char)
//	Return value: 1: OK / 0: NG EEPROM NOACK(unsigned char)
//----------------------------------------------------------------------
unsigned char Write_byte_EEPROM(unsigned short adrs, unsigned char wr_data)
{
	unsigned char	loop,ack;

	ack = 0;
	loop = 0;
	while (loop <= WR_RETRY_CNT){
		_IIC._ICCR.BYTE = 0x89;	// ICE=1(P57,P56->SCL,SDA) 
		_IIC._ICMR.BYTE = 0x08;
		_TSCR.BYTE = 0xfc;
		while (_IIC._ICCR.BIT.BBSY != 0);	// Bus busy? 
		ack = Set_adrs_EEPROM(adrs);
		if (ack==1){
			loop = WR_OK;
		}else{
			_IIC._ICCR.BIT.IRIC = 0;
			_IIC._ICCR.BYTE = _IIC._ICCR.BYTE & 0xfa;	// Generate stop condition
			wait_eeprom(1000);	// tWC=1ms 
			loop++;
		}
	}
	if (ack==1){
		ack = Write_data_EEPROM(wr_data);
		_IIC._ICCR.BIT.IRIC = 0;
		_IIC._ICCR.BYTE = _IIC._ICCR.BYTE & 0xfa;		// Generate stop condition 
	}
	return ack;
}

